# -*- coding: utf-8 -*-
"""
triton
~~~~~~~~

:copyright: (c) 2015 by Postmates
"""

__title__ = 'triton'
__version__ = '0.0.19'
__description__ = 'Triton - Kinesis Data Pipeline'
__url__ = 'https://github.com/postmates/triton-python'
__build__ = 0
__author__ = 'Postmates, Inc'
__author_email__ = 'rhettg@gmail.com'
__license__ = 'ISC'
__copyright__ = 'Copyright 2015 Postmates'

from .config import load_config
from .stream import get_stream
from .nonblocking_stream import get_nonblocking_stream
from .store import stream_from_s3_store



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
import logging
import psycopg2.pool
import time

from triton import errors

log = logging.getLogger(__name__)


triton_dsn = os.environ.get('TRITON_DB')
triton_client_name = os.environ.get('TRITON_CLIENT_NAME')
postal_rds_pool = None

CREATE_TABLE_STMT = """
CREATE TABLE IF NOT EXISTS triton_checkpoint (
    client VARCHAR(255) NOT NULL,
    stream VARCHAR(255) NOT NULL,
    shard VARCHAR(255) NOT NULL,
    seq_num VARCHAR(255) NOT NULL,
    updated INTEGER NOT NULL,
    PRIMARY KEY (client, stream, shard))
"""


def get_triton_connection_pool():
    global postal_rds_pool
    if postal_rds_pool is None:
        if not triton_dsn:
            raise errors.TritonCheckpointError(
                'triton_dsn not configured with TRITON_DB evn variable')
        postal_rds_pool = psycopg2.pool.ThreadedConnectionPool(
            1, 20, triton_dsn
        )
    return postal_rds_pool


def init_db(db_pool_function=get_triton_connection_pool):
    """Create the required table in DB if not already there"""
    db_pool = db_pool_function()
    conn = db_pool.getconn()
    curs = conn.cursor()
    curs.execute(CREATE_TABLE_STMT)
    conn.commit()


class TritonCheckpointer(object):
    """Handles checkpoints for triton"""

    def __init__(self, stream_name, client_name=triton_client_name):
        if not client_name:
            raise errors.TritonCheckpointError(
                'client_name is required to create a TritonCheckpointer')
        self.client_name = client_name
        self.stream_name = stream_name
        self.db_pool = get_triton_connection_pool()

        self.checkpoint_exists_sql = (
            "SELECT 1 FROM triton_checkpoint WHERE client=%s "
            "AND stream=%s AND shard=%s"
        )

        self.update_checkpoint_sql = (
            "UPDATE triton_checkpoint SET seq_num=%s, updated=%s"
            " WHERE client=%s AND stream=%s AND shard=%s"
        )

        self.create_checkpoint_sql = (
            "INSERT INTO triton_checkpoint VALUES (%s, %s, %s, %s, %s)"
        )

        self.last_seq_no_sql = (
            "SELECT seq_num FROM triton_checkpoint WHERE "
            "client=%s AND stream=%s AND shard=%s"
        )

    def checkpoint(self, shard_id, sequence_number):
        conn = self.db_pool.getconn()
        try:
            checkpoint_exists = True
            curs = conn.cursor()
            curs.execute(
                self.checkpoint_exists_sql,
                (self.client_name, self.stream_name, shard_id)
            )
            try:
                prev_sequence_number = curs.fetchone()
                log.info('checkpoint: {}'.format(prev_sequence_number))
                if prev_sequence_number is None:
                    checkpoint_exists = False
            except psycopg2.ProgrammingError:
                checkpoint_exists = False

            if checkpoint_exists:
                log.info('Updating checkpoint for {}-{}: {}'.format(
                    self.stream_name, shard_id, sequence_number))
                curs.execute(
                    self.update_checkpoint_sql,
                    (
                        str(sequence_number), time.time(),
                        self.client_name, self.stream_name, shard_id
                    )
                )
            else:
                log.info('creating checkpoint for {}-{}: {}'.format(
                    self.stream_name, shard_id, sequence_number))
                curs.execute(
                    self.create_checkpoint_sql,
                    (
                        self.client_name, self.stream_name, shard_id,
                        str(sequence_number), time.time(),
                    )
                )
        except Exception:
            conn.rollback()
            raise
        else:
            conn.commit()
        finally:
            self.db_pool.putconn(conn)

    def last_sequence_number(self, shard_id):
        conn = self.db_pool.getconn()
        try:
            log.info('getting last_sequence_number for {}-{}'.format(
                self.stream_name, shard_id))
            curs = conn.cursor()
            curs.execute(
                self.last_seq_no_sql,
                (self.client_name, self.stream_name, shard_id)
            )
            try:
                sequence_number = curs.fetchone()
                if sequence_number is None:
                    return  # no checkpoint, we're done
                return sequence_number[0]
            except psycopg2.ProgrammingError:
                return  # no checkpoint, we're done
        except Exception:
            conn.rollback()
            raise
        else:
            conn.commit()
        finally:
            self.db_pool.putconn(conn)



from __future__ import unicode_literals
import decimal
import datetime


def msgpack_encode_default(obj):
    """Extra encodings for python types into msgpack

    These are attempted if our normal serialization fails.
    """
    if isinstance(obj, decimal.Decimal):
        return str(obj)
    if isinstance(obj, datetime.datetime):
        return obj.isoformat(str(' '))
    if isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")
    if hasattr(obj, 'coords'):
        # hack to deal with lat-long points
        return repr(obj.coords)
    try:
        return repr(obj)
    except Exception:
        raise TypeError("Unknown type: %r" % (obj,))

def unicode_to_ascii_str(text):
    #if unicode, escape out multibyte characters
    if isinstance(text, unicode):
        return text.encode('utf-8')
    else:
        #need to str here because this function could be fed something
        #that's not unicode but needs to be an ascii string (e.g. an int)
        return str(text)

def ascii_to_unicode_str(text):
    #if ascii/escaped unicode, decode to utf-8
    if isinstance(text, str):
        return text.decode('utf-8')
    else:
        #need to unicode here because this function could be fed something
        #that's not an ascii str but needs to be a unicode string (e.g. an int)
        return unicode(text)



from __future__ import unicode_literals
import time
import io
import datetime
import os.path
import re
import logging
import itertools

import msgpack
import snappy
import boto.s3
from boto.s3.connection import OrdinaryCallingFormat
from encoding import ascii_to_unicode_str, unicode_to_ascii_str

MAX_BUFFER_SIZE = 1024 * 1024

log = logging.getLogger(__name__)


class StreamArchiveWriter(object):

    def __init__(self, stream_config, base_dt, base_path):
        self.config = stream_config
        self.base_dt = base_dt
        self.ts = time.time()
        self.base_path = base_path

        self.buffer = io.BytesIO()
        self.writer = None

    @property
    def file_path(self):
        date_str = self.base_dt.strftime('%Y%m%d')
        file_name = "{}-archive-{}.tri".format(ascii_to_unicode_str(self.config['name']),
                                               int(self.ts))
        return os.path.join(self.base_path, date_str, file_name)

    def put(self, **kwargs):
        data = msgpack.packb(kwargs)

        self.buffer.write(data)

        if self.buffer.tell() >= MAX_BUFFER_SIZE:
            self.flush()

    def flush(self):
        if self.buffer.tell() == 0:
            return

        if self.writer is None:
            try:
                os.makedirs(os.path.dirname(self.file_path))
            except OSError:
                pass

            self.writer = io.open(unicode_to_ascii_str(self.file_path), mode="wb")
            self.snappy_compressor = snappy.StreamCompressor()

        data = self.snappy_compressor.add_chunk(self.buffer.getvalue())
        self.writer.write(data)

        # Reset our buffer
        self.buffer.truncate(0)
        self.buffer.seek(0)

    def close(self):
        self.flush()

        if self.writer:
            self.writer.close()
            self.writer = None


def decoder(stream):
    """Generator that processes data from the stream (by iterating) and yields
    triton records"""
    snappy_stream = snappy.StreamDecompressor()
    unpacker = msgpack.Unpacker(encoding='utf-8')
    for data in stream:
        buf = snappy_stream.decompress(data)
        if buf:
            unpacker.feed(buf)
            # Oh to have yield from
            for rec in unpacker:
                yield rec


class StreamArchiveReader(object):

    def __init__(self, file_path):
        self.file_path = file_path

    def __iter__(self):
        f = io.open(unicode_to_ascii_str(self.file_path), mode="rb")
        return decoder(f)


class ArchiveFile(object):
    """Represents a stream archive stored in S3 file

    This mostly translates between file names and the properties that we can
    infer from the names.
    """

    def __init__(self, stream_name, day, ts, shard=None):
        self.stream_name = stream_name
        self.day = day
        self.ts = ts
        self.shard = shard
        self.is_archive = bool(shard is None)

    @property
    def file_path(self):
        return "{}/{}-{}-{}.tri".format(
            self.day.strftime('%Y%m%d'), self.stream_name,
            self.shard or 'archive', self.ts)

    def s3_key(self, bucket):
        return boto.s3.key.Key(bucket, name=self.file_path)

    def open(self, bucket):
        """Create a iterable stream of data from the log file.
        """
        return decoder(self.s3_key(bucket))

    @classmethod
    def from_s3_key(cls, key):
        match = re.match(r"(?P<day>\d{8})\/(?P<stream>.+)\-(?P<ts>\d+)\.tri$",
                         key.name)
        if match is None:
            raise ValueError(key)

        shard = None
        match_info = match.groupdict()
        if match_info['stream'].endswith('-archive'):
            stream_name = match_info['stream'][:-len('-archive')]
        else:
            stream_name = match_info['stream'][:match_info['stream'].find(
                '-shardId')]
            shard = re.match(r".*(shardId-\d+)$", match_info['stream']).group(1)
            if not shard:
                raise ValueError(stream_name)

        ts = int(match_info['ts'])
        dt = datetime.datetime.strptime(match_info['day'], '%Y%m%d')
        return cls(stream_name, dt.date(), ts, shard)


def inclusive_date_range(start_dt, end_dt):
    start_date = start_dt.date()
    end_date = end_dt.date()

    current_date = start_date
    while current_date <= end_date:
        yield current_date
        current_date += datetime.timedelta(days=1)


def s3_prefix_for_date_and_name(date, stream_name):
    date_str = date.strftime('%Y%m%d')
    return "{}/{}-".format(date_str, stream_name)


def find_log_files_in_s3(bucket, stream_name, start_dt, end_dt):
    prefixes = []
    for dt in inclusive_date_range(start_dt, end_dt):
        prefixes.append(s3_prefix_for_date_and_name(dt, stream_name))

    all_files = []
    for pf in prefixes:
        day_files = []
        for key in bucket.list(pf):
            try:
                af = ArchiveFile.from_s3_key(key)
                # If the file is marked as an archive, it's the only file we
                # should pay attention to. It represents a full dump of the day.
                if af.is_archive:
                    day_files = [af]
                    break

                day_files.append(af)
            except ValueError as e:
                log.warning("S3 key %r not a log file: %r", key, e)
                continue

        day_files.sort(key=lambda f: f.ts)
        all_files += day_files

    return all_files


def open_bucket(bucket_name, region_name):
    # We need to specify calling_format to get around bugs in having a '.' in
    # the bucket name
    conn = boto.s3.connect_to_region(region_name,
                                     calling_format=OrdinaryCallingFormat())

    return conn.get_bucket(bucket_name)


def stream_from_s3_store(bucket, stream_config, start_dt, end_dt):
    log_files = find_log_files_in_s3(
        bucket, stream_config['name'], start_dt, end_dt)

    streams = []
    for lf in log_files:
        data_stream = lf.open(bucket)
        streams.append(data_stream)

    return itertools.chain(*streams)



# -*- coding: utf-8 -*-
"""
triton.nonblocking_stream
~~~~~~~~

This module provides for a way to write to a Triton Stream without
blocking to wait for a successful write to the stream.
It does this by sending the messageover ZeroMQ to tritond.

Adapted from https://github.com/rhettg/BlueOx/blob/master/blueox/network.py

"""
from __future__ import unicode_literals
import logging
import threading
import struct
import atexit

import zmq
import msgpack

from triton import errors
from triton import config
from triton.encoding import msgpack_encode_default, unicode_to_ascii_str, ascii_to_unicode_str

log = logging.getLogger(__name__)

# We want to limit how many messages we'll hold in memory so if our tritond is
# unavailable, we don't just run out of memory.  I based this value on rough
# value of rather large 3k sized messages, and how many we can fit in 10 megs.
MAX_QUEUED_MESSAGES = 3500

# If we have pending outgoing messages, this is how long we'll wait after
# being told to exit.
LINGER_SHUTDOWN_MSECS = 3000

# version byte in our meta struct for future upgrading.
META_STRUCT_VERSION = 0x4
META_STRUCT_FMT = "!B64p64p"

threadLocal = threading.local()

# Context can be shared between threads
_zmq_context = None
_connect_str = None


def init(host, port):
    global _zmq_context
    global _connect_str

    _zmq_context = zmq.Context()
    _connect_str = "tcp://%s:%d" % (host, port)


def _thread_connect():
    if _zmq_context and not getattr(threadLocal, 'zmq_socket', None):
        threadLocal.zmq_socket = _zmq_context.socket(zmq.PUSH)
        threadLocal.zmq_socket.hwm = MAX_QUEUED_MESSAGES
        threadLocal.zmq_socket.linger = LINGER_SHUTDOWN_MSECS

        threadLocal.zmq_socket.connect(_connect_str)


class NonblockingStream(object):

    def __init__(self, name, partition_key):
        self.name = name
        if len(self.name) > 64:
            raise ValueError("Stream Name Too Long")
        self.partition_key = partition_key
        if _zmq_context is None:
            init(*config.get_zmq_config())

    def _partition_key(self, data):
        return ascii_to_unicode_str(data[self.partition_key])

    def _serialize_context(self, data):
        # Our sending format is made up of two messages. The first has a
        # quick to unpack set of meta data that our collector is going to
        # use for routing and stats. This is much faster than having the
        # collector decode the whole event. We're just going to use python
        # struct module to make a quick and dirty data structure
        if len(self._partition_key(data)) > 64:
            raise ValueError("Partition Key Too Long")

        meta_data = struct.pack(META_STRUCT_FMT, META_STRUCT_VERSION,
                                unicode_to_ascii_str(self.name),
                                unicode_to_ascii_str(self._partition_key(data)))
        try:
            message_data = msgpack.packb(data)
        except TypeError:
            # If we fail to serialize our context, we can try again with an
            # enhanced packer (it's slower though)
            message_data = msgpack.packb(data, default=msgpack_encode_default)

        return meta_data, message_data

    def put(self, **kwargs):
        global _zmq_context
        _thread_connect()

        try:
            meta_data, message_data = self._serialize_context(kwargs)
        except Exception:
            log.exception(
                "Triton serialization failure for stream {}".format(
                    self.name))
            return

        if _zmq_context and threadLocal.zmq_socket is not None:
            try:
                threadLocal.zmq_socket.send_multipart(
                    (meta_data, message_data), zmq.NOBLOCK)
            except zmq.ZMQError:
                log.exception(
                    ("Failed sending Triton event over ZMQ. ",
                        "Buffer full? tritond not running?"))
        else:
            log.info("Skipping sending event %s", self.name)


def get_nonblocking_stream(stream_name, config):
    s_config = config.get(stream_name)
    if not s_config:
        raise errors.StreamNotConfiguredError()

    return NonblockingStream(stream_name, s_config['partition_key'])


def close():
    global _zmq_context

    if getattr(threadLocal, 'zmq_socket', None):
        threadLocal.zmq_socket.close()
        threadLocal.zmq_socket = None

    _zmq_context = None

atexit.register(close)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
"""
This module contains the set of triton's exceptions

:copyright: (c) 2012 by Firstname Lastname.
:license: ISC, see LICENSE for more details.

"""


class Error(Exception):
    """This is an ambiguous error that occured."""
    pass


class TritonNotConfiguredError(Error):
    """Indicates no config file found"""
    pass


class StreamNotConfiguredError(Error):
    pass


class ShardNotFoundError(Error):
    """Indicates the requested shard isn't known"""
    pass


class EndOfShardError(Error):
    pass


class KinesisError(Error):
    """An ambiguous or unknown Kinesis Error"""
    pass


class KinesisPutManyError(Error):
    """An ambiguous or unknown Kinesis Error"""
    def __init__(self, reason, failed_data=None, *args):
        super(KinesisPutManyError, self).__init__(reason, *args)
        self.failed_data = failed_data


class TritonCheckpointError(Error):
    """Error of misconfiguration of Triton Checkpointing"""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import glob
from distutils.core import setup

PACKAGES = ['triton']


def get_init_val(val, packages=PACKAGES):
    pkg_init = "%s/__init__.py" % PACKAGES[0]
    value = '__%s__' % val
    fn = open(pkg_init)
    for line in fn.readlines():
        if line.startswith(value):
            return line.split('=')[1].strip().strip("'")


setup(
    name='py-%s' % get_init_val('title'),
    version=get_init_val('version'),
    description=get_init_val('description'),
    long_description=open('README.md').read(),
    author=get_init_val('author'),
    author_email=get_init_val('author_email'),
    url=get_init_val('url'),
    scripts=glob.glob("bin/triton*"),
    install_requires=['python-snappy', 'msgpack_python', 'pyzmq', 'future'],
    packages=PACKAGES
)



from __future__ import unicode_literals
import base64
import time
import logging

import msgpack
import boto.kinesis.layer1
from boto.kinesis.exceptions import ProvisionedThroughputExceededException
from boto.exception import BotoServerError
import boto.regioninfo

from triton import errors
from triton.checkpoint import TritonCheckpointer
from triton.encoding import msgpack_encode_default, unicode_to_ascii_str, ascii_to_unicode_str

import pystatsd

MIN_POLL_INTERVAL_SECS = 1.0
KINESIS_MAX_LENGTH = 500  # Can't write more than 500 records at a time
KINESIS_MAX_RETRYS = 2  # Kinesis 'InternalFailure' retry attempts

ITER_TYPE_LATEST = 'LATEST'
ITER_TYPE_ALL = 'TRIM_HORIZON'
ITER_TYPE_FROM_SEQNUM = 'AFTER_SEQUENCE_NUMBER'
ITER_TYPE_FROM_CHECKPOINT = 'FROM_CHECKPOINT'

log = logging.getLogger(__name__)

# See https://github.com/postmates/pystatsd for statsd configuration
# Note that pystatsd writes raise no exception if no statsd server is running
STATSD_PREFIX = 'triton.stream.'
STATSD_PUT_ATTEMPT = STATSD_PREFIX + "put_attempt."


class Record(object):
    __slots__ = ['shard_id', 'seq_num', 'data']

    def __init__(self, shard_id, seq_num, data):
        self.shard_id = shard_id
        self.seq_num = seq_num
        self.data = data

    @classmethod
    def _decode_record_data(cls, record_data):
        return msgpack.unpackb(base64.b64decode(record_data), encoding='utf-8')

    @classmethod
    def from_raw_record(cls, shard_id, raw_record):
        return cls(shard_id, raw_record['SequenceNumber'],
                   cls._decode_record_data(raw_record['Data']))

    def __repr__(self):
        return u'<Record {} {}>'.format(self.shard_id, self.seq_num)


class StreamIterator(object):
    """Handles the workflow of reading from a shard

    Args:
        stream - Instance of Stream()
        shard_id - Which shard we're reading from (shardId-00001)
        iterator_type - How we're iterating, like 'LATEST' or 'TRIM_HORIZON'
        seq_num - For 'AFTER_SEQUENCE_NUMBER' type, tells us where to start.
        fallback_iterator_type - For 'AFTER_SEQUENCE_NUMBER', if there is no
            checkpoint availible, create an iterator with this iterator_type
            instead
    """

    def __init__(
        self, stream, shard_id, iterator_type,
        seq_num=None, fallback_iterator_type=ITER_TYPE_ALL
    ):
        self.stream = stream
        self.shard_id = shard_id
        self.iterator_type = iterator_type
        self.seq_num = seq_num
        self.fallback_iterator_type = fallback_iterator_type

        self._iter_value = None
        self.records = []

        self._empty = True
        self.behind_latest_secs = None

        self._checkpointer = None
        self.last_seq_num = None

    @property
    def checkpointer(self):
        if self._checkpointer is None:
            self._checkpointer = TritonCheckpointer(self.stream.name)
        return self._checkpointer

    def checkpoint(self):
        if self.last_seq_num is not None:
            self.checkpointer.checkpoint(self.shard_id, self.last_seq_num)

    @property
    def iter_value(self):
        if self._iter_value is None:
            if self.iterator_type == ITER_TYPE_FROM_CHECKPOINT:
                self.seq_num = self.checkpointer.last_sequence_number(
                    self.shard_id)
                if self.seq_num is None:
                    self.iterator_type = self.fallback_iterator_type
                else:
                    self.iterator_type = ITER_TYPE_FROM_SEQNUM
            log.info(
                "Creating iterator %r", (
                    self.stream.name, self.shard_id,
                    self.iterator_type, self.seq_num))
            i = self.stream.conn.get_shard_iterator(
                self.stream.name, self.shard_id, self.iterator_type,
                self.seq_num)
            self._iter_value = i['ShardIterator']

        return self._iter_value

    def fill(self):
        try:
            record_resp = self.stream.conn.get_records(self.iter_value,
                                                       b64_decode=False)
        except ProvisionedThroughputExceededException:
            # We set our poll interval to be conservative (and match
            # recommended kinesis libraries) but it's always possible that
            # insufficient capacity has been provisioned. Most likely, this is
            # transient and something we can recover from. But we should
            # complain loudly.
            log.error("Rate exceeded for %r:%r", self.stream.name,
                      self.shard_id)
            return

        behind_latest_secs = record_resp['MillisBehindLatest'] / 1000.0
        log.debug("Found %d records filling %r (behind %d secs)",
                  len(record_resp['Records']), self, int(behind_latest_secs))

        if self.behind_latest_secs > 0.0 and behind_latest_secs == 0:
            log.info("%r has caught up with latest", self)

        if self.behind_latest_secs is None:
            log.info("%r behind latest by %ds", self, behind_latest_secs)

        self.behind_latest_secs = behind_latest_secs

        for rec in record_resp['Records']:
            self.records.append(Record.from_raw_record(self.shard_id, rec))
            self._empty = False

        if record_resp.get('NextShardIterator'):
            self._iter_value = record_resp['NextShardIterator']
        else:
            # If a next iterator isn't provided, it probably indicates the
            # shard has be ended due to split or merge.
            raise errors.EndOfShardError()

    def __iter__(self):
        return self

    def next(self):
        if self._empty:
            self.fill()

        try:
            rec = self.records.pop(0)
            self.last_seq_num = rec.seq_num
            return rec

        except IndexError:
            self._empty = True
            raise StopIteration

    def __repr__(self):
        return u'<StreamIterator {} {} ({})>'.format(
            self.stream.name, self.shard_id, self.iterator_type)


class CombinedStreamIterator(object):
    """Combines multiple StreamIterators for reading from multiple shards

    Handles load balancing between streams.
    """

    def __init__(self, iterators):
        self.iterators = iterators
        self._fill_iterators = set()
        self._running = True
        self._last_wait = None

        self.last_iterator = None
        self.last_seq_num = None

        self._records = []

    def _wait(self):
        if self._last_wait is None:
            # First fill, no waiting
            self._last_wait = time.time()
            return

        secs_since_fill = time.time() - self._last_wait

        throttle_secs = MIN_POLL_INTERVAL_SECS - secs_since_fill
        if throttle_secs > 0.0:
            throttle_secs = MIN_POLL_INTERVAL_SECS - secs_since_fill
            log.debug("Throttling for %f secs", throttle_secs)
            time.sleep(throttle_secs)

        self._last_wait = time.time()

    def _fill(self):
        if not self._fill_iterators:
            self._wait()
            self._fill_iterators.update(set(self.iterators))

        iter_to_fill = self._fill_iterators.pop()
        self.last_iterator = iter_to_fill
        log.debug("Checking stream (%s, %s) ", iter_to_fill.stream.name,
                  iter_to_fill.shard_id)
        self._records += list(iter_to_fill)

    def __iter__(self):
        return self

    def next(self):
        # The goals is simple:
        # 1. Deliver any record already loaded before loading any new ones.
        # 2. Don't starve any streams
        # 3. Don't hammer empty shards
        # 4. Don't load more records after stop() is called
        while True:
            try:
                rec = self._records.pop(0)
                self.last_seq_num = rec.seq_num
                return rec
            except IndexError:
                if not self._running:
                    raise StopIteration

                self._fill()

    def stop(self):
        self._running = False

    def checkpoint(self):
        for this_iterator in set(self.iterators):
            if this_iterator != self.last_iterator:
                # we've already processed all data pulled from this iterator
                this_iterator.checkpoint()
            else:
                # we could be in the middle of processing this data
                # checkpoint only as far as the data retrieved via next()
                this_iterator.checkpointer.checkpoint(
                    this_iterator.shard_id, self.last_seq_num)


class Stream(object):

    def __init__(self, conn, name, partition_key):
        self.conn = conn
        self.name = ascii_to_unicode_str(name)
        self.partition_key = ascii_to_unicode_str(partition_key)
        self._shard_ids = None

    #NOTE: explanation of the convoluted try blocks in _partition_key!
    #when looking up the partition_key in the data, we need to first check
    #the unicode version of the key, then check the escaped/ascii version.
    #If the key truly is missing, we want to swallow the second KeyError and
    #return the first KeyError that points to self.partition_key and not the
    #copy created by calling unicode_to_ascii_str
    def _partition_key(self, data):
        try:
            return ascii_to_unicode_str(data[self.partition_key])
        except KeyError as original_error:
            #supports a user putting escaped unicode data in, still returns
            #unicode partition key
            try:
                return ascii_to_unicode_str(data[unicode_to_ascii_str(self.partition_key)])
            except KeyError:
                pass
            raise original_error

    @property
    def shard_ids(self):
        if self._shard_ids is None:
            self._shard_ids = []
            stream_describe_resp = self.conn.describe_stream(self.name)

            if stream_describe_resp['StreamDescription']['HasMoreShards']:
                raise NotImplementedError

            for shard in stream_describe_resp['StreamDescription']['Shards']:
                self._shard_ids.append(shard['ShardId'])

        return self._shard_ids

    def _select_shard_ids(self, shard_nums):
        shard_ids = []
        if shard_nums:
            for shard_num in shard_nums:
                try:
                    shard_ids.append(self.shard_ids[shard_num])
                except IndexError:
                    raise errors.ShardNotFoundError()
        else:
            shard_ids = self.shard_ids

        return shard_ids

    def put(self, **kwargs):
        pystatsd.increment(
            STATSD_PUT_ATTEMPT + self.name
        )

        try:
            data = msgpack.packb(kwargs)
        except TypeError:
            # If we fail to serialize our context, we can try again with an
            # enhanced packer (it's slower though)
            data = msgpack.packb(kwargs, default=msgpack_encode_default)
        resp = _call_and_retry(
            self.conn.put_record,
            self.name, data,
            self._partition_key(kwargs)
        )

        try:
            return resp['ShardId'], resp['SequenceNumber']
        except KeyError:
            raise errors.KinesisError((
                'An unknown error occurred for stream {},'
                ' response was {}').format(self.name, resp))

    def put_many(self, records):
        data_recs = []
        for r in records:
            try:
                data = msgpack.packb(r)
            except TypeError:
                # If we fail to serialize our context, we can try again with an
                # enhanced packer (it's slower though)
                data = msgpack.packb(r, default=msgpack_encode_default)
            data_recs.append({
                'Data': data,
                'PartitionKey': self._partition_key(r),
            })

        return self._put_many_packed(data_recs)

    def _put_many_packed(self, records, retry_count=0, b64_encode=True):
        """Re-usable method for already packed messages,
            used here and by tritond for non-blocking writes

        args:
            records - list() of dicts with the following structure:
                {
                    'Data': msgpacked-data,
                    'PartitionKey': partition_key of the record
                }
            retry_count - number of retries for individual failed records
            b64_encode  - parameter to boto kinesis library included b/c boto
                          changes the data record itself. We need to pass false
                          to prevent re-encoding on failure.
        """
        resp_value = []
        num_records = len(records)
        max_record = 0
        retry_records = []

        pystatsd.increment(
            STATSD_PUT_ATTEMPT + self.name,
            num_records
        )

        while max_record < num_records:
            # iterate through all the records MAX_LENTH records at a time

            # Note that the following _call_and_retry will only
            # retry for 500 server errors;
            # ProvisionedThroughputExceededException
            # will not happen for put_records
            resp = _call_and_retry(
                self.conn.put_records,
                records[max_record:max_record + KINESIS_MAX_LENGTH],
                self.name,
                b64_encode=b64_encode)

            for idx, r in enumerate(resp['Records']):
                # Per http://docs.aws.amazon.com/kinesis/
                # ...latest/APIReference/API_PutRecordsResultEntry.html
                # ProvisionedThroughputExceededException and InternalFailure
                # happen on a message by message basis.
                # Check for individual failed messages and queue for retry
                try:
                    resp_value.append((r['ShardId'], r['SequenceNumber']))
                except KeyError:
                    retry_records.append(records[max_record + idx])
            max_record += KINESIS_MAX_LENGTH
        if retry_records:
            # if any individual messages have failed, retry them now
            if retry_count > KINESIS_MAX_RETRYS:
                raise errors.KinesisPutManyError(
                    'Failed to put_many records to Kinesis',
                    failed_data=retry_records)
            else:
                time.sleep(2 ** retry_count * .1)
                resp_value.extend(self._put_many_packed(
                    retry_records,
                    retry_count=retry_count + 1,
                    b64_encode=False)
                )
        return resp_value

    def build_iterator_for_all(self, shard_nums=None):
        shard_ids = self._select_shard_ids(shard_nums)
        return self._build_iterator(ITER_TYPE_ALL, shard_ids, None)

    def build_iterator_from_seqnum(self, shard_id, seq_num):
        return self._build_iterator(ITER_TYPE_FROM_SEQNUM, [shard_id], seq_num)

    def build_iterator_from_latest(self, shard_nums=None):
        shard_ids = self._select_shard_ids(shard_nums)
        return self._build_iterator(ITER_TYPE_LATEST, shard_ids, None)

    def build_iterator_from_checkpoint(self, shard_nums=None):
        shard_ids = self._select_shard_ids(shard_nums)
        return self._build_iterator(ITER_TYPE_FROM_CHECKPOINT, shard_ids, None)

    def _build_iterator(self, iterator_type, shard_ids, seq_num):
        all_iters = []
        for shard_id in shard_ids:
            i = StreamIterator(self, shard_id, iterator_type, seq_num)
            all_iters.append(i)

        return CombinedStreamIterator(all_iters)


def connect_to_region(region_name, **kw_params):
    # NOTE(rhettg): current version of boto doesn't know about us-west-1 for
    # kinesis
    region = boto.regioninfo.RegionInfo(
        name=region_name,
        endpoint='kinesis.{}.amazonaws.com'.format(region_name),
        connection_cls=boto.kinesis.layer1.KinesisConnection)

    return region.connect(**kw_params)


def get_stream(stream_name, config):
    s_config = config.get(stream_name)
    if not s_config:
        raise errors.StreamNotConfiguredError()

    conn = connect_to_region(s_config.get('region', 'us-east-1'))

    return Stream(conn, s_config['name'], s_config['partition_key'])


def _call_and_retry(kinesis_function, *args, **kwargs):
    """
    Retry Logic for generic kinesis calls.

    This code follows the exponetial backoff pattern suggested by
    http://docs.aws.amazon.com/general/latest/gr/api-retries.html
    """
    retries = 0
    while True:
        try:
            return kinesis_function(*args, **kwargs)
        except BotoServerError as e:
            if retries >= KINESIS_MAX_RETRYS:
                raise e
            if (
                    e.status / 100 == 5  # 5xx error
                    or
                    type(e) == ProvisionedThroughputExceededException
            ):
                time.sleep(2 ** retries * .1)
                retries += 1
            else:
                raise e



from __future__ import unicode_literals
import io
import yaml
import logging
import os

from . import errors

ENV_VAR_TRITON_ZMQ_HOST = 'TRITON_ZMQ_HOST'
ENV_VAR_TRITON_ZMQ_PORT = 'TRITON_ZMQ_PORT'
ZMQ_DEFAULT_HOST = '127.0.0.1'
ZMQ_DEFAULT_PORT = 3515

log = logging.getLogger(__name__)

REQUIRED_CONFIG_KEYS = ['name', 'partition_key']

_zmq_config = None

#NOTE: when loading config dictionary, yaml automatically converts unicode
#in the yaml file to unicode in the dictionary. Dictionary will still contain
#ascii keys and values for ascii strings in the yaml file
def load_config(file_name):
    try:
        config_dict = yaml.load(io.open(file_name))
    except IOError as e:
        log.error("Failed to open %s: %r", file_name, e)
        return None

    for stream_name, v in config_dict.iteritems():
        for k in REQUIRED_CONFIG_KEYS:
            if k not in v:
                raise errors.InvalidConfigurationError(
                    "Missing {} : {}".format(stream_name, k))

    return config_dict


def get_zmq_config():
    global _zmq_config
    if not _zmq_config:
        zmq_host = os.environ.get(ENV_VAR_TRITON_ZMQ_HOST, ZMQ_DEFAULT_HOST)
        zmq_port = int(os.environ.get(ENV_VAR_TRITON_ZMQ_PORT,
                                      ZMQ_DEFAULT_PORT))
        _zmq_config = (zmq_host, zmq_port)

    return _zmq_config
